#include <vcl.h>
#include <process.h>
#pragma hdrstop
#include "AnimTimer.h"

USEPACKAGE("vcl35.bpi");
USEUNIT("AnimTimer.cpp");
USERES("AnimTimer.dcr");

//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------

namespace Animtimer
{
  class TThreadMiser;
  enum TThreadMessageCode {tmcHadEnough=1, tmcThreadTimer=2, tmcAlertThread=3};
  typedef struct {
    HWND hwnd; // set at thread creation
    CRITICAL_SECTION critsec; // critsec protects IsHwndvalid
    bool IsHwndValid; // set to False if the window no longer wants the thread
    HANDLE hChangeEvent; // in response to a SCRM_CHANGE, to interrupt the sleep
    unsigned long idThread; // so, if it's there, it can be told to quit
    unsigned long hThread; // a handle so we can wait for it
    TThreadMiser *miser;
  } TThreadData;
}
// The meaning of TThreadData:
// critsec protects both fields. isHwndValid says whether or not it's
// valid. Obviously, if it isn't, then messages can't be sent to it.
// Normally 'IsHwndValid' will be set to TRUE on thread creation, and set
// to FALSE just before a PostThreadMessage(WM_QUIT) has been sent to the
// thread. IsHwndValid will only ever switch from TRUE to FALSE.
// A TThreadData is created externally before thread creation, and it is
// passed to the thread function, and it is the responsibility of the thread
// function to delete it after having received a WM_QUIT message.
// We use this peculiar protocol just in case the thread spins a few more
// times before it finally gets the WM_QUIT message.
// If ever a thread decides it has had enough, it posts a SCRM_threadhadenough
// message to the main window, and will expect the main window to send it the
// WM_QUIT.


//---------------------------------------------------------------------------
// We use the ThreadMiser component, who will be owned by the TApplication,
// because TApplication's destructor will be executed after the end of the
// message loop; and at that time it will be safe to wait for the threads

namespace Animtimer
{
class TThreadMiser : public TComponent
{
public:
  HANDLE FHandle; TAnimTimer *FTimer;
  __fastcall TThreadMiser(TComponent *AOwner);
  __fastcall ~TThreadMiser();
  __property HANDLE Handle = {read=FHandle,write=FHandle};
  __property TAnimTimer* AnimTimer={read=FTimer,write=FTimer};
};
}

__fastcall Animtimer::TThreadMiser::TThreadMiser(TComponent *AOwner):TComponent(AOwner)
{ FHandle=NULL; FTimer=NULL; }
__fastcall Animtimer::TThreadMiser::~TThreadMiser()
{ if (AnimTimer!=NULL)
  { //ScrDebug(tlMsg,"ThreadMiser"+AnsiString((int)this)+": deleting the AnimTimer first.");
    delete AnimTimer; AnimTimer=NULL;
  }
  if (FHandle!=NULL)
  { DWORD res=WaitForSingleObject(FHandle,5000);
    if (res==WAIT_TIMEOUT)
    { //ScrDebug(tlErr,"ThreadMiser"+AnsiString((int)this)+": Not all threads have terminated: "+AnsiString((int)FHandle));
    }
    CloseHandle(FHandle);
    FHandle=NULL;
  }
}

void TAnimTimer::ThreadFunction(void *lParam)
{ Animtimer::TThreadData *td=(Animtimer::TThreadData*)lParam;
  //ScrDebug(tlMsg,"Thread func "+AnsiString((int)td->hThread)+": starting. +++");
  HWND hwnd=td->hwnd;
  bool mustquit=false;
  bool isvalid;
  bool hadenough=false;
  // isvalid is the signal from outside; hadenough is the signal from inside
  MSG msg;
  while (!mustquit)
  { // before anything, check if hwnd is valid
    bool res;
    // If we've had enough, then we know that we don't want to do anything more
    // and we are just waiting for the quit message so we can leave
    if (hadenough)
    { //ScrDebug(tlMsg,"Thread "+AnsiString((int)td->hThread)+" has had enough and is waiting to quit.");
      res=true; GetMessage(&msg,NULL,0,0xFFFF);
    }
    else
    { res=PeekMessage(&msg,NULL,0,0xFFFF,PM_REMOVE); }
    if (res)
    { if (msg.message==WM_QUIT)
      { mustquit=true;
        //ScrDebug(tlMsg,"Thread "+AnsiString((int)td->hThread)+" has got a quit message.");
      }
      else
      { //ScrDebug(tlMsg,"Thread "+AnsiString((int)td->hThread)+" message: msg="+AnsiString((int)msg.message)+" wParam="+AnsiString((int)msg.wParam)+" lParam="+AnsiString((int)msg.lParam));
      }
    }
    if (!mustquit && !hadenough)
    { DWORD starttime,period,endtime;
      EnterCriticalSection(&td->critsec);
      isvalid=td->IsHwndValid;
      LeaveCriticalSection(&td->critsec);
      if (isvalid)
      { starttime=GetTickCount();
        period=SendMessage(hwnd,WM_APP,Animtimer::tmcThreadTimer,0);
        endtime=GetTickCount();
      }
      if (isvalid)
      { switch (period)
        { case -1: break; // not even a sleep
          case -2: // had enough.
          { hadenough=true;
            PostMessage(hwnd,WM_APP,Animtimer::tmcHadEnough,0);
          } break;
          default:
          { DWORD sleeptime=0;
            if (starttime+period>endtime) sleeptime=starttime+period-endtime;
            if (sleeptime==0) Sleep(0);
            else
            { DWORD res=WaitForSingleObject(td->hChangeEvent,sleeptime);
              if (res==WAIT_OBJECT_0) ResetEvent(td->hChangeEvent);
            }
          }
        }
      }
    }
  }
  CloseHandle(td->hChangeEvent);
  td->miser->Handle=NULL;
  Application->RemoveComponent(td->miser);
  delete td;
  //ScrDebug(tlMsg,"Thread "+AnsiString((int)td->hThread)+" coming to an end. Deleted thread-data. +++");
}



__fastcall TAnimTimer::TAnimTimer(TComponent* Owner): TComponent(Owner)
{ IsMultiThread=true;
  hwnd=NULL;
  if (!ComponentState.Contains(csDesigning)) hwnd=AllocateHWnd(this->LuWndProc);
  FEnabled=true;
  FInterval=10;
  UpdateTimer();
}
__fastcall TAnimTimer::~TAnimTimer()
{ FEnabled=false;
  UpdateTimer();
  if (hwnd!=NULL) DeallocateHWnd(hwnd); hwnd=NULL;
}
void __fastcall TAnimTimer::Timer()
{ if (FOnTimer!=NULL) FOnTimer(this);
}
void __fastcall TAnimTimer::SetEnabled(bool newenabled)
{ if (newenabled==FEnabled) return;
  FEnabled=newenabled;
  UpdateTimer();
}
void __fastcall TAnimTimer::SetInterval(unsigned int newinterval)
{ if (FInterval==newinterval) return;
  FInterval=newinterval;
  Alert();
}
void __fastcall TAnimTimer::SetOnTimer(TNotifyEvent val)
{ FOnTimer=val;
  UpdateTimer();
}
void __fastcall TAnimTimer::UpdateTimer()
{ if (hwnd==NULL) return;
  Stop();
  if (FEnabled && FOnTimer!=NULL) Start();
}
void __fastcall TAnimTimer::LuWndProc(TMessage &Message)
{ Message.Result=DefWindowProc(hwnd,Message.Msg,Message.WParam,Message.LParam);
  switch (Message.Msg)
  { case WM_APP:
    { switch (Message.WParam)
      { case Animtimer::tmcHadEnough:
        { //ScrDebug(tlMsg,"Timer "+Name+" SCR_HADENOUGH: closing down thread...");
          Animtimer::TThreadData *td=(Animtimer::TThreadData*)GetWindowLong(hwnd,GWL_USERDATA);
          if (td==NULL)
          { //ScrDebug(tlErr,"Timer "+Name+" SCR_HADENOUGH: strangely, the timer had disappeared!");
          }
          else
          { EnterCriticalSection(&td->critsec);
            td->IsHwndValid=false;
            LeaveCriticalSection(&td->critsec);
            PostThreadMessage(td->idThread,WM_QUIT,0,0);
          }
          SetWindowLong(hwnd,GWL_USERDATA,0);
        } break;
        case Animtimer::tmcThreadTimer:
        { Animtimer::TThreadData *td=(Animtimer::TThreadData *)GetWindowLong(hwnd,GWL_USERDATA);
          if (td!=NULL) Timer();
          Message.Result=FInterval;
        } break;
        case Animtimer::tmcAlertThread:
        { Animtimer::TThreadData *td=(Animtimer::TThreadData *)GetWindowLong(hwnd,GWL_USERDATA);
          if (td!=NULL) SetEvent(td->hChangeEvent);
        } break;
      }
    } break;
  }
}
void __fastcall TAnimTimer::Start()
{ Animtimer::TThreadData *td;
  if (hwnd==NULL)
  { //ScrDebug(tlMsg,"AnimTimer "+Name+" start: cannot start, because no hwnd.");
    return;
  }
  td=(Animtimer::TThreadData*)GetWindowLong(hwnd,GWL_USERDATA);
  if (td!=NULL)
  { //ScrDebug(tlMsg,"AnimTimer "+Name+" start: cannot start, because a timer already there.");
    return;
  }
  td=new Animtimer::TThreadData;
  SetWindowLong(hwnd,GWL_USERDATA,(LONG)td);
  InitializeCriticalSection(&td->critsec);
  td->hChangeEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
  td->hwnd=hwnd;
  td->IsHwndValid=TRUE;
  HANDLE holdThread;
  holdThread=(HANDLE)_beginthreadNT(ThreadFunction,4096,td,NULL,0,&(td->idThread));
  HANDLE hThread;
  DuplicateHandle(GetCurrentProcess(),holdThread,GetCurrentProcess(),&hThread,0,TRUE,DUPLICATE_SAME_ACCESS);
  td->hThread=(unsigned long)hThread;
  Animtimer::TThreadMiser *miser=new Animtimer::TThreadMiser(Application);
  miser->Handle=(HANDLE)td->hThread;
  miser->AnimTimer=this;
  td->miser=miser;
  HANDLE hThisT=GetCurrentThread();
  HANDLE hThatT=(HANDLE)td->hThread;
  int pr=GetThreadPriority(hThisT);
  SetThreadPriority(hThatT,pr-1);
  // Hopefully it won't matter if we underflow by one
  //ScrDebug(tlMsg,"AnimTimer "+Name+" "+AnsiString((int)td->hThread)+" has been created! ***");
}
void __fastcall TAnimTimer::Stop()
{ if (hwnd==NULL)
  { //ScrDebug(tlErr,"AnimTimer "+Name+" stop: but hwnd is null!");
    return;
  }
  Animtimer::TThreadData *td=(Animtimer::TThreadData*)GetWindowLong(hwnd,GWL_USERDATA);
  if (td==NULL)
  { // trying to stop one that didn't exist anyway. This is not necessarily bad.
    return;
  }
  //ScrDebug(tlMsg,"AnimTimer "+Name+" stop: destroying timer...");
  EnterCriticalSection(&td->critsec);
  td->IsHwndValid=false;
  LeaveCriticalSection(&td->critsec);
  // That will be enough to ensure that we enver get any messages
  // from now on. Note that the thread's send(SCRM_TIMER) is itself bracketed
  // inside this critical section
  // so not even another SendMessage can get through
  PostThreadMessage(td->idThread,WM_QUIT,0,0);
  SetWindowLong(hwnd,GWL_USERDATA,0);
  //ScrDebug(tlMsg,"AnimTimer "+Name+" stop: we have left thread "+AnsiString((int)td->hThread)+" to its own devices. ***");
}
void __fastcall TAnimTimer::Alert()
{ if (hwnd==NULL)
  { //ScrDebug(tlErr,"AnimTimer "+Name+" alert: but cannot alert hwnd, since no hwnd!");
    return;
  }
  SendMessage(hwnd,WM_APP,Animtimer::tmcAlertThread,0);
}



//---------------------------------------------------------------------------
namespace Animtimer
{ void __fastcall PACKAGE Register()
  { TComponentClass classes[1] = {__classid(TAnimTimer)};
    RegisterComponents("ScrPlus", classes, 0);
  }
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void*)
{
        return 1;
}
//---------------------------------------------------------------------------
